
package edu.umn.cs.kumarbio.shear;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Random;
import java.util.Scanner;
import java.util.TreeSet;
import java.util.concurrent.ExecutionException;

import edu.umn.cs.kumarbio.Genome;
import edu.umn.cs.kumarbio.UnexpectedBehaviorException;

public class ShearSV {

  private String  prefix;
  private String  origBamFile;
  private String  origFaRefFile;
  private String  origBwaIndexPrefix;
  private String  origTwoBitRefFile;
  private String  region;
  private String  crestDir;
  private boolean svOnly;
  private boolean debug;
  private String  rgId;
  private String  rgLb;
  private String  rgPl;
  private String  rgPu;
  private String  rgSm;

  /**
   * Creates a new ShearSV object by initializing the local variables that are
   * read from the command line arguments.
   * 
   * @param args
   *        The command line arguments to process, excluding the first one
   *        (SHEAR command, i.e. 'sv' or 'assemble')
   */
  public ShearSV(String[] args) throws ShearUsageException {

    // Read in arguments
    debug = false;
    crestDir = System.getenv("CREST_DIRECTORY");
    for (int i = 0; i < args.length; i++) {
      if (args[i].equals("-p")) {
        prefix = args[++i];
      } else if (args[i].equals("-b")) {
        origBamFile = args[++i];
      } else if (args[i].equals("-f")) {
        origFaRefFile = args[++i];
      } else if (args[i].equals("-i")) {
        origBwaIndexPrefix = args[++i];
      } else if (args[i].equals("-t")) {
        origTwoBitRefFile = args[++i];
      } else if (args[i].equals("-r")) {
        region = args[++i];
      } else if (args[i].equals("-d")) {
        debug = true;
      } else if (args[i].equals("--sv-only")) {
        svOnly = true;
      } else {
        throw new ShearUsageException("Unknown argument: " + args[i]);
      }
    }

    // Validate prefix
    if (prefix == null) {
      throw new ShearUsageException("Must provide a prefix with the '-p' argument");
    }
    if (!prefix.matches("^[a-zA-Z0-9_\\-\\.]+$")) {
      throw new ShearUsageException("Prefix must consist of alphanumeric characters, underscores, dashes, or periods.");
    }

    // Validate origBamFile
    if (origBamFile == null) {
      throw new ShearUsageException("Must provide a BAM alignment file with the '-b' argument");
    }
    if (!(new File(origBamFile).exists())) {
      throw new ShearUsageException("BAM alignment file does not exist: " + origBamFile);
    } else if (!(new File(origBamFile + ".bai").exists())) {
      throw new ShearUsageException("BAM alignment index file does not exist: " + origBamFile + ".bai");
    } else {
      origBamFile = new File(origBamFile).getAbsolutePath();
    }

    // Validate origFaRefFile
    if (origFaRefFile == null) {
      throw new ShearUsageException("Must provide a FASTA reference file with the '-f' argument");
    }
    if (!(new File(origFaRefFile).exists())) {
      throw new ShearUsageException("FASTA reference file does not exist: " + origFaRefFile);
    } else if (!(new File(origFaRefFile + ".fai").exists())) {
      throw new ShearUsageException("FASTA reference index file does not exist: " + origFaRefFile + ".fai");
    } else {
      origFaRefFile = new File(origFaRefFile).getAbsolutePath();
    }

    // Validate origBwaIndexPrefix
    if (origBwaIndexPrefix == null) {
      origBwaIndexPrefix = origFaRefFile;
    }
    if (!(new File(origBwaIndexPrefix + ".bwt").exists())) {
      throw new ShearUsageException("BWA index prefix appears to be invalid: " + origBwaIndexPrefix);
    } else {
      origBwaIndexPrefix = new File(origBwaIndexPrefix + ".bwt").getAbsolutePath();
      origBwaIndexPrefix = origBwaIndexPrefix.substring(0, origBwaIndexPrefix.length() - 4);
    }

    // Validate origTwoBitRefFile
    if (origTwoBitRefFile == null) {
      throw new ShearUsageException("Must provide a Two-Bit reference file with the '-t' argument");
    }
    if (!(new File(origTwoBitRefFile).exists())) {
      throw new ShearUsageException("Two-Bit reference file does not exist: " + origTwoBitRefFile);
    } else {
      origTwoBitRefFile = new File(origTwoBitRefFile).getAbsolutePath();
    }

    // Validate region
    if ((region != null) && !region.matches("[a-zA-Z0-9]+(:[0-9,]+\\-[0-9,]+)?")) {
      throw new ShearUsageException("Invalid region specified: " + region);
    }

    // Validate crestDir
    if (crestDir == null) {
      throw new ShearUsageException("Must specify the CREST directory via the CREST_DIRECTORY environment variable");
    }
    if (!(new File(crestDir).exists())) {
      throw new ShearUsageException("CREST directory does not exist: " + crestDir);
    } else {
      crestDir = new File(crestDir).getAbsolutePath();
    }
    if (!(new File(crestDir + "/extractSClip.pl").exists())) {
      throw new ShearUsageException("CREST directory does not contain the necessary script extractSClip.pl: "
          + crestDir);
    }
    if (!(new File(crestDir + "/CREST.pl").exists())) {
      throw new ShearUsageException("CREST directory does not contain the necessary script CREST.pl: " + crestDir);
    }

    // Output settings information
    System.out.println("[SHEAR] Starting SHEAR-SV");
    System.out.println("[SHEAR] Prefix:                 " + prefix);
    System.out.println("[SHEAR] BAM alignment file:     " + origBamFile);
    System.out.println("[SHEAR] FASTA reference file:   " + origFaRefFile);
    System.out.println("[SHEAR] BWA index prefix:       " + origBwaIndexPrefix);
    System.out.println("[SHEAR] Two-Bit reference file: " + origTwoBitRefFile);
    System.out.println("[SHEAR] CREST directory:        " + crestDir);
    if (region != null) {
      System.out.println("[SHEAR] Region:                 " + region);
    }
    if (svOnly) {
      System.out.println("[SHEAR] Prediction mode:        SV Only");
    } else {
      System.out.println("[SHEAR] Prediction mode:        All");
    }
    if (debug) {
      System.out.println("[SHEAR] Debugging mode:         ON");
    } else {
      System.out.println("[SHEAR] Debugging mode:         OFF");
    }

  }

  private ArrayList<CrestPrediction> getCrestPreds(String crestPredFile) throws FileNotFoundException {
    ArrayList<CrestPrediction> preds = new ArrayList<CrestPrediction>();
    Scanner sc = new Scanner(new File(crestPredFile));
    while (sc.hasNextLine()) {
      String predLine = sc.nextLine();
      String[] tokens = predLine.split("\t");
      // What about cross-chromosome events?
      String bp1region = tokens[0];
      int bp1 = Integer.parseInt(tokens[1]);
      String bp1dir = tokens[2];
      String bp2region = tokens[4];
      int bp2 = Integer.parseInt(tokens[5]);
      String bp2dir = tokens[6];
      String type = tokens[8];
      preds.add(new CrestPrediction(type, bp1region, bp1, bp1dir, 0, 0, bp2region, bp2, bp2dir, 0, 0, 0));
    }
    sc.close();
    return preds;
  }

  private ArrayList<CrestPrediction> filterCrestPreds(ArrayList<CrestPrediction> origPreds) {
    // SHOULD MAKE SURE SORTED
    // NEED TO CHECK FOR SVs WITH SAME LOC ON BOTH BREAKPOINTS
    ArrayList<CrestPrediction> newPreds = new ArrayList<CrestPrediction>();
    tryNextPred:
    for (int i = 0; i < origPreds.size(); i++) {
      CrestPrediction origPred = origPreds.get(i);
      for (int k = 0; k < newPreds.size(); k++) {
        CrestPrediction newPred = newPreds.get(k);
        if ((origPred.bp1 == newPred.bp1) && origPred.bp1region.equals(newPred.bp1region)) {
          if (Math.abs(origPred.bp2 - newPred.bp2) <= 10) {
            if (newPred.bp2sc < origPred.bp2sc) {
              newPreds.set(k, origPred);
            }
            continue tryNextPred;
          }
        } else if ((origPred.bp2 == newPred.bp2) && origPred.bp2region.equals(newPred.bp2region)) {
          if (Math.abs(origPred.bp1 - newPred.bp1) <= 10) {
            if (newPred.bp1sc < origPred.bp1sc) {
              newPreds.set(k, origPred);
            }
            continue tryNextPred;
          }
        }
      }
      newPreds.add(origPreds.get(i));
    }
    return newPreds;
  }

  private void outputResults(TreeSet<Variant> variants) throws IOException {

    Genome gen = null;
    PrintWriter allSdiOut = null;
    PrintWriter allReportOut = null;
    PrintWriter snpIndelSdiOut = null;
    PrintWriter snpIndelReportOut = null;
    PrintWriter svSdiOut = null;
    PrintWriter svReportOut = null;

    try {

      gen = new Genome(origFaRefFile);
      if (!svOnly) {
        allSdiOut = new PrintWriter(prefix + ".all.sdi");
        allReportOut = new PrintWriter(prefix + ".all.report");
        snpIndelSdiOut = new PrintWriter(prefix + ".snpindel.sdi");
        snpIndelReportOut = new PrintWriter(prefix + ".snpindel.report");
      }
      svSdiOut = new PrintWriter(prefix + ".sv.sdi");
      svReportOut = new PrintWriter(prefix + ".sv.report");

      for (Variant v : variants) {
        // Fix report format to account for chromosome name lengths
        String sdiLine = null;
        String reportLine = null;

        // Output SNP
        if (v instanceof SNP) {
          SNP snp = (SNP) v;
          sdiLine = getSdiLine(snp.getChr(), snp.getLoc(), Character.toString(snp.getRefBase()),
              Character.toString(snp.getVarBase()));
          reportLine = String.format("%-8s%-30s%7.2f%%", "SNP", snp.getChr() + ":" + snp.getLoc(),
              snp.getHeterogeneity());

          // Output Deletion
        } else if (v instanceof Deletion) {
          Deletion del = (Deletion) v;
          sdiLine = getSdiLine(del.getChr(), del.getLoc(),
              gen.getSequence(del.getChr(), del.getLoc(), del.getEndLoc()), "-");
          reportLine = String.format("%-8s%-30s%7.2f%%", "DEL",
              del.getChr() + ":" + del.getLoc() + "-" + del.getEndLoc(), del.getHeterogeneity());

          // Output Insertion
        } else if (v instanceof Insertion) {
          Insertion ins = (Insertion) v;
          sdiLine = getSdiLine(ins.getChr(), ins.getLoc(), "-", ins.getInsertedBases());
          reportLine = String.format("%-8s%-30s%7.2f%%", "INS",
              ins.getChr() + ":" + ins.getLoc() + " (" + ins.getSize() + " bp)", ins.getHeterogeneity());

          // Output TandemDuplication
        } else if (v instanceof TandemDuplication) {
          TandemDuplication tdup = (TandemDuplication) v;
          sdiLine = getSdiLine(tdup.getChr(), tdup.getLoc(), "-",
              gen.getSequence(tdup.getChr(), tdup.getLoc(), tdup.getEndLoc()));
          reportLine = String.format("%-8s%-30s%7.2f%%", "TDUP",
              tdup.getChr() + ":" + tdup.getLoc() + "-" + tdup.getEndLoc(), tdup.getHeterogeneity());

          // Output Inversion
        } else if (v instanceof Inversion) {
          Inversion inv = (Inversion) v;
          sdiLine = getSdiLine(inv.getChr(), inv.getLoc(),
              gen.getSequence(inv.getChr(), inv.getLoc(), inv.getEndLoc()),
              Genome.rc(gen.getSequence(inv.getChr(), inv.getLoc(), inv.getEndLoc())));
          reportLine = String.format("%-8s%-30s%7.2f%%", "INV",
              inv.getChr() + ":" + inv.getLoc() + "-" + inv.getEndLoc(), inv.getHeterogeneity());

          // Output Translocation
        } else if (v instanceof Translocation) {
          Translocation xloc = (Translocation) v;
          reportLine = String.format("%-8s%-30s%7.2f%%", "XLOC",
              xloc.getChr() + ":" + xloc.getLoc() + "-" + xloc.getEndChr() + ":" + xloc.getEndLoc(),
              xloc.getHeterogeneity());

          // Unknown variant type
        } else {
          throw new IllegalArgumentException("Encountered an invalid SV type while outputting results");
        }

        if (!svOnly) {
          allSdiOut.print(sdiLine);
          allReportOut.println(reportLine);
        }
        System.out.println("[SHEAR]     " + reportLine);
        if (!svOnly && ((v instanceof SNP) || v.isIndel())) {
          snpIndelSdiOut.print(sdiLine);
          snpIndelReportOut.println(reportLine);
        } else {
          svSdiOut.print(sdiLine);
          svReportOut.println(reportLine);
        }

      }

    } finally {
      gen.close();
      if (!svOnly) {
        allSdiOut.close();
        allReportOut.close();
        snpIndelSdiOut.close();
        snpIndelReportOut.close();
      }
      svSdiOut.close();
      svReportOut.close();
    }

  }

  private ArrayList<CrestPrediction> processCrestPreds(String bamFileName, String refFileName,
      ArrayList<CrestPrediction> crestPreds) throws Exception {
    for (int i = 0; i < crestPreds.size(); i++) {
      CrestPrediction pred = crestPreds.get(i);
      System.out.printf("[SHEAR] Processing: %s %s:%d(%s) - %s:%d(%s)\n", pred.type, pred.bp1region, pred.bp1,
          pred.bp1dir, pred.bp2region, pred.bp2, pred.bp2dir);
      if (pred.type.equals("DEL")) {
        crestPreds.set(i, processDel(bamFileName, refFileName, pred));
      } else if (pred.type.equals("INS")) {
        crestPreds.set(i, processIns(bamFileName, refFileName, pred));
      } else if (pred.type.equals("ITX")) {
        crestPreds.set(i, processItx(bamFileName, refFileName, pred));
      } else if (pred.type.equals("CTX")) {
        crestPreds.set(i, processCtx(bamFileName, refFileName, pred));
      } else if (pred.type.equals("INV")) {
        crestPreds.set(i, processInv(bamFileName, refFileName, pred));
      } else {
        throw new IllegalArgumentException("The following SV type is not valid: " + pred.type);
      }
    }
    return crestPreds;
  }

  private CrestPrediction processDel(String bamFileName, String refFileName, CrestPrediction pred) throws Exception {

    Random rand = new Random(System.currentTimeMillis());
    String tmpFile1 = "tmp-" + pred.bp1region + pred.bp1 + "-" + rand.nextInt(Integer.MAX_VALUE);
    String tmpFile2 = "tmp-" + pred.bp2region + pred.bp2 + "-" + rand.nextInt(Integer.MAX_VALUE);

    String[] gatkArgs1 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp1), "-clipDirection", "+", "-L", pred.bp1region + ":" + pred.bp1, "-o", tmpFile1, "-l",
        "ERROR" };
    if (debug) {
      gatkArgs1[gatkArgs1.length - 1] = "INFO";
    }
    RunGatkCommand.run(gatkArgs1, debug);

    String[] gatkArgs2 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp2), "-clipDirection", "-", "-L", pred.bp2region + ":" + pred.bp2, "-o", tmpFile2, "-l",
        "ERROR" };
    if (debug) {
      gatkArgs2[gatkArgs2.length - 1] = "INFO";
    }
    RunGatkCommand.run(gatkArgs2, debug);

    File bp1file = new File(tmpFile1);
    Scanner bp1scanner = new Scanner(bp1file);
    int bp1sc = bp1scanner.nextInt();
    int bp1span = bp1scanner.nextInt();
    bp1scanner.close();
    bp1file.delete();

    File bp2file = new File(tmpFile2);
    Scanner bp2scanner = new Scanner(bp2file);
    int bp2sc = bp2scanner.nextInt();
    int bp2span = bp2scanner.nextInt();
    bp2scanner.close();
    bp2file.delete();

    double per = (100 * ((double) bp1sc + (double) bp2sc))
        / ((double) bp1sc + (double) bp2sc + (((double) bp1span + (double) bp2span) / 2));
    if (per > 100) {
      per = 100;
    }
    pred.bp1sc = bp1sc;
    pred.bp1span = bp1span;
    pred.bp2sc = bp2sc;
    pred.bp2span = bp2span;
    pred.heterogeneityPercent = per;

    System.out.println(String.format("[SHEAR]     Estimated heterogeneity:                %.2f%%",
        pred.heterogeneityPercent));

    return pred;

  }

  private CrestPrediction processIns(String bamFileName, String refFileName, CrestPrediction pred) throws Exception {

    Random rand = new Random(System.currentTimeMillis());
    String tmpFile1 = "tmp-" + pred.bp1region + pred.bp1 + "-" + rand.nextInt(Integer.MAX_VALUE);
    String tmpFile2 = "tmp-" + pred.bp2region + pred.bp2 + "-" + rand.nextInt(Integer.MAX_VALUE);

    String[] gatkArgs1 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp2), "-clipDirection", "-", "-L", pred.bp2region + ":" + pred.bp2, "-o", tmpFile1, "-l",
        "ERROR" };
    if (debug) {
      gatkArgs1[gatkArgs1.length - 1] = "INFO";
    }
    RunGatkCommand.run(gatkArgs1, debug);

    String[] gatkArgs2 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp1), "-clipDirection", "+", "-L", pred.bp1region + ":" + pred.bp1, "-o", tmpFile2, "-l",
        "ERROR" };
    if (debug) {
      gatkArgs2[gatkArgs2.length - 1] = "INFO";
    }
    RunGatkCommand.run(gatkArgs2, debug);

    File bp1file = new File(tmpFile1);
    Scanner bp1scanner = new Scanner(bp1file);
    int bp1sc = bp1scanner.nextInt();
    int bp1span = bp1scanner.nextInt();
    bp1scanner.close();
    bp1file.delete();

    File bp2file = new File(tmpFile2);
    Scanner bp2scanner = new Scanner(bp2file);
    int bp2sc = bp2scanner.nextInt();
    int bp2span = bp2scanner.nextInt();
    bp2scanner.close();
    bp2file.delete();

    double per = (100 * ((double) bp1sc + (double) bp2sc)) / (((double) bp1span + (double) bp2span) / 2);
    if (per > 100) {
      per = 100;
    }
    pred.bp1sc = bp1sc;
    pred.bp1span = bp1span;
    pred.bp2sc = bp2sc;
    pred.bp2span = bp2span;
    pred.heterogeneityPercent = per;

    System.out.println(String.format("[SHEAR]     Estimated heterogeneity:                %.2f%%",
        pred.heterogeneityPercent));

    return pred;

  }

  private CrestPrediction processItx(String bamFileName, String refFileName, CrestPrediction pred) throws Exception {

    // FIX clip directions

    Random rand = new Random(System.currentTimeMillis());
    String tmpFile1 = "tmp-" + pred.bp1region + pred.bp1 + "-" + rand.nextInt(Integer.MAX_VALUE);
    String tmpFile2 = "tmp-" + pred.bp2region + pred.bp2 + "-" + rand.nextInt(Integer.MAX_VALUE);

    String[] gatkArgs1 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp1), "-clipDirection", "+", "-L", pred.bp1region + ":" + pred.bp1, "-o", tmpFile1, "-l",
        "ERROR" };
    if (debug) {
      gatkArgs1[gatkArgs1.length - 1] = "INFO";
    }
    RunGatkCommand.run(gatkArgs1, debug);

    String[] gatkArgs2 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp2), "-clipDirection", "-", "-L", pred.bp2region + ":" + pred.bp2, "-o", tmpFile2, "-l",
        "ERROR" };
    if (debug) {
      gatkArgs2[gatkArgs2.length - 1] = "INFO";
    }
    RunGatkCommand.run(gatkArgs2, debug);

    File bp1file = new File(tmpFile1);
    Scanner bp1scanner = new Scanner(bp1file);
    int bp1sc = bp1scanner.nextInt();
    int bp1span = bp1scanner.nextInt();
    bp1scanner.close();
    bp1file.delete();

    File bp2file = new File(tmpFile2);
    Scanner bp2scanner = new Scanner(bp2file);
    int bp2sc = bp2scanner.nextInt();
    int bp2span = bp2scanner.nextInt();
    bp2scanner.close();
    bp2file.delete();

    double per = (100 * ((double) bp1sc + (double) bp2sc)) / ((double) bp1sc + (double) bp2sc + bp1span + bp2span);
    if (per > 100) {
      per = 100;
    }
    pred.bp1sc = bp1sc;
    pred.bp1span = bp1span;
    pred.bp2sc = bp2sc;
    pred.bp2span = bp2span;
    pred.heterogeneityPercent = per;

    System.out.println(String.format("[SHEAR]     Estimated heterogeneity:                %.2f%%",
        pred.heterogeneityPercent));

    return pred;

  }

  private CrestPrediction processCtx(String bamFileName, String refFileName, CrestPrediction pred) throws Exception {

    // FIX clip directions

    Random rand = new Random(System.currentTimeMillis());
    String tmpFile1 = "tmp-" + pred.bp1region + pred.bp1 + "-" + rand.nextInt(Integer.MAX_VALUE);
    String tmpFile2 = "tmp-" + pred.bp2region + pred.bp2 + "-" + rand.nextInt(Integer.MAX_VALUE);

    String[] gatkArgs1 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp1), "-clipDirection", "+", "-L", pred.bp1region + ":" + pred.bp1, "-o", tmpFile1, "-l",
        "ERROR" };
    if (debug) {
      gatkArgs1[gatkArgs1.length - 1] = "INFO";
    }
    RunGatkCommand.run(gatkArgs1, debug);

    String[] gatkArgs2 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp2), "-clipDirection", "-", "-L", pred.bp2region + ":" + pred.bp2, "-o", tmpFile2, "-l",
        "ERROR" };
    if (debug) {
      gatkArgs2[gatkArgs2.length - 1] = "INFO";
    }
    RunGatkCommand.run(gatkArgs2, debug);

    File bp1file = new File(tmpFile1);
    Scanner bp1scanner = new Scanner(bp1file);
    int bp1sc = bp1scanner.nextInt();
    int bp1span = bp1scanner.nextInt();
    bp1scanner.close();
    bp1file.delete();

    File bp2file = new File(tmpFile2);
    Scanner bp2scanner = new Scanner(bp2file);
    int bp2sc = bp2scanner.nextInt();
    int bp2span = bp2scanner.nextInt();
    bp2scanner.close();
    bp2file.delete();

    double per = (100 * ((double) bp1sc + (double) bp2sc)) / ((double) bp1sc + (double) bp2sc + bp1span + bp2span);
    if (per > 100) {
      per = 100;
    }
    pred.bp1sc = bp1sc;
    pred.bp1span = bp1span;
    pred.bp2sc = bp2sc;
    pred.bp2span = bp2span;
    pred.heterogeneityPercent = per;

    System.out.println(String.format("[SHEAR]     Estimated heterogeneity:                %.2f%%",
        pred.heterogeneityPercent));

    return pred;

  }

  private CrestPrediction processInv(String bamFileName, String refFileName, CrestPrediction pred) throws Exception {

    Random rand = new Random(System.currentTimeMillis());
    String tmpFile1 = "tmp-" + pred.bp1region + pred.bp1 + "-" + rand.nextInt(Integer.MAX_VALUE);
    String tmpFile2 = "tmp-" + pred.bp2region + pred.bp2 + "-" + rand.nextInt(Integer.MAX_VALUE);

    String[] gatkArgs1 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp1), "-clipDirection", "+", "-unclippedSpanningOnly", "-L",
        pred.bp1region + ":" + pred.bp1, "-o", tmpFile1, "-l", "ERROR" };
    if (debug) {
      gatkArgs1[gatkArgs1.length - 1] = "INFO";
    }
    RunGatkCommand.run(gatkArgs1, debug);

    String[] gatkArgs2 = { "-T", "Heterogeneity", "-I", bamFileName, "-R", refFileName, "-clipLocus",
        String.valueOf(pred.bp2), "-clipDirection", "-", "-unclippedSpanningOnly", "-L",
        pred.bp2region + ":" + pred.bp2, "-o", tmpFile2, "-l", "ERROR" };
    if (debug) {
      gatkArgs2[gatkArgs2.length - 1] = "INFO";
    }
    RunGatkCommand.run(gatkArgs2, debug);

    File bp1file = new File(tmpFile1);
    Scanner bp1scanner = new Scanner(bp1file);
    int bp1sc = bp1scanner.nextInt();
    int bp1span = bp1scanner.nextInt();
    bp1scanner.close();
    bp1file.delete();

    File bp2file = new File(tmpFile2);
    Scanner bp2scanner = new Scanner(bp2file);
    int bp2sc = bp2scanner.nextInt();
    int bp2span = bp2scanner.nextInt();
    bp2scanner.close();
    bp2file.delete();

    double per = (100 * (2d * ((double) bp1sc + (double) bp2sc)))
        / ((2d * ((double) bp1sc + (double) bp2sc)) + bp1span + bp2span);
    if (per > 100) {
      per = 100;
    }
    pred.bp1sc = bp1sc;
    pred.bp1span = bp1span;
    pred.bp2sc = bp2sc;
    pred.bp2span = bp2span;
    pred.heterogeneityPercent = per;

    System.out.println(String.format("[SHEAR]     Estimated heterogeneity:                %.2f%%",
        pred.heterogeneityPercent));

    return pred;

  }

  /*
   * Returns the properly formatted SDI line for a given region name, start
   * location, reference string, and consensus string.
   */
  public static String getSdiLine(String regionName, int startLoc, String reference, String consensus) {
    StringBuffer line = new StringBuffer();
    line.append(regionName);
    line.append("\t");
    line.append(startLoc);
    line.append("\t");
    if (reference.equals("-")) {
      if (consensus.equals("-")) {
        line.append(0);
      } else {
        line.append(consensus.length());
      }
    } else if (consensus.equals("-")) {
      if (reference.equals("-")) {
        line.append(0);
      } else {
        line.append("-");
        line.append(reference.length());
      }
    } else {
      line.append(consensus.length() - reference.length());
    }
    line.append("\t");
    line.append(reference);
    line.append("\t");
    line.append(consensus);
    line.append("\t*\t*\n");
    return line.toString();
  }

  private void saveReadGroupInfo(String bamFile) throws InterruptedException, IOException, ExecutionException,
      Exception {
    String tmpOutput = prefix + ".tmp" + System.currentTimeMillis();
    FileOutputStream output = new FileOutputStream(tmpOutput + ".origHeaders");
    if (debug) {
      RunSystemCommand.run(new ProcessBuilder("samtools", "view", "-H", bamFile), output, null, System.out, "samtools",
          debug);
    } else {
      RunSystemCommand.run(new ProcessBuilder("samtools", "view", "-H", bamFile), output, null, null, null, debug);
    }
    output.close();
    Scanner sc = new Scanner(new File(tmpOutput + ".origHeaders"));
    rgId = prefix;
    rgLb = prefix;
    rgPl = prefix;
    rgPu = prefix;
    rgSm = prefix;
    while (sc.hasNextLine()) {
      String line = sc.nextLine();
      if (line.startsWith("@RG")) {
        String[] tokens = line.split("\t");
        for (int i = 1; i < tokens.length; i++) {
          if (tokens[i].startsWith("ID")) {
            rgId = tokens[i].substring(3);
          } else if (tokens[i].startsWith("LB")) {
            rgLb = tokens[i].substring(3);
          } else if (tokens[i].startsWith("PL")) {
            rgPl = tokens[i].substring(3);
          } else if (tokens[i].startsWith("PU")) {
            rgPu = tokens[i].substring(3);
          } else if (tokens[i].startsWith("SM")) {
            rgSm = tokens[i].substring(3);
          }
        }
        if (rgLb == null) {
          rgLb = rgId;
        }
        if (rgPl == null) {
          rgPl = rgId;
        }
        if (rgPu == null) {
          rgPu = rgId;
        }
        if (rgSm == null) {
          rgSm = rgId;
        }
        break;
      }
    }
    sc.close();
    new File(tmpOutput + ".origHeaders").delete();
    // Add the read group info back, in case it was missing before
    addReadGroups(bamFile, prefix + "-orig-with-rg.bam");
    if (debug) {
      RunSystemCommand.run(new ProcessBuilder("samtools", "index", prefix + "-orig-with-rg.bam"), System.out,
          "samtools", System.out, "samtools", debug);
    } else {
      RunSystemCommand.run(new ProcessBuilder("samtools", "index", prefix + "-orig-with-rg.bam"), null, null, null,
          null, debug);
    }
  }

  private TreeSet<Variant> runHaplotypeCaller(String bamFile, String faRefFile, String vcfOutputFile) throws Exception {
    String[] gatkArgs = { "-T", "HaplotypeCaller", "-I", bamFile, "-R", faRefFile, "-o", vcfOutputFile, "-l", "ERROR" };
    if (debug) {
      gatkArgs[gatkArgs.length - 1] = "INFO";
    }
    RunGatkCommand.run(gatkArgs, debug);
    Scanner sc = new Scanner(new File(vcfOutputFile));
    TreeSet<Variant> variants = new TreeSet<Variant>();
    while (sc.hasNextLine()) {
      String line = sc.nextLine();
      if (line.charAt(0) != '#') {
        String[] tokens = line.split("\t");
        String chr = tokens[0];
        int loc = Integer.valueOf(tokens[1]);
        String refBases = tokens[3];
        String varBases = tokens[4];
        int variantSize = varBases.length() - refBases.length();
        Variant v = null;
        if (refBases.matches("^[ACGT]+$") && varBases.matches("^[ACGT]+$")) {
          if ((variantSize == 0) & (refBases.length() == 1)) {
            v = new SNP(chr, loc, refBases.charAt(0), varBases.charAt(0));
          } else if ((variantSize > 0) & (refBases.length() == 1) & (refBases.charAt(0) == varBases.charAt(0))) {
            // Check for weird case VCF-locs (ends of starts of chr)
            v = new Insertion(chr, loc + 1, varBases.substring(1));
            v.setIndel(true);
          } else if ((variantSize < 0) & (varBases.length() == 1) & (refBases.charAt(0) == varBases.charAt(0))) {
            // Check for weird case VCF-locs (ends of starts of chr)
            v = new Deletion(chr, loc + 1, -variantSize);
            v.setIndel(true);
          } else {
            if (debug) {
              System.out.println("[SHEAR] Skipping unexpected event in HaplotypeCaller output (" + chr + ":" + loc
                  + " " + refBases + " -> " + varBases + ")");
            }
            continue;
          }
        } else {
          if (debug) {
            System.out.println("[SHEAR] Skipping unexpected event in HaplotypeCaller output (" + chr + ":" + loc + " "
                + refBases + " -> " + varBases + ")");
          }
          continue;
        }
        String[] formatTags = tokens[8].split(":");
        String[] formatInfo = tokens[9].split(":");
        double heterogeneity = Double.NaN;
        for (int i = 0; i < formatTags.length; i++) {
          if (formatTags[i].equals("AD")) {
            String[] alleleDepths = formatInfo[i].split(",");
            double varDepth = Double.valueOf(alleleDepths[1]);
            double refDepth = Double.valueOf(alleleDepths[0]);
            heterogeneity = 100.0 * (varDepth / (varDepth + refDepth));
          }
        }
        if (heterogeneity != Double.NaN) {
          if (heterogeneity > 0) {
            v.setHeterogeneity(heterogeneity);
            variants.add(v);
          } else if (debug) {
            System.out.println("[SHEAR] Filtering out SNP/INDEL (" + chr + ":" + loc + " " + refBases + " -> "
                + varBases + ") with 0% heterogeneity, considered false positive");
          }
        } else {
          sc.close();
          throw new UnexpectedBehaviorException("AD not found or not containing valid data");
        }
      }
    }
    sc.close();
    return variants;
  }

  private void runCrest(String bamFileName, String crestPrefix, String faRefFile, GFServer gfServer)
      throws InterruptedException, IOException, ExecutionException {
    ProcessBuilder extractSClipProcess = new ProcessBuilder("perl", "-I" + crestDir, crestDir + "/extractSClip.pl",
        "-p", crestPrefix, "-i", bamFileName, "--ref_genome", faRefFile);
    ProcessBuilder crestProcess = new ProcessBuilder("perl", "-I" + crestDir, crestDir + "/CREST.pl", "-max_rep_cover",
        "1000000", "-p", crestPrefix, "-f", crestPrefix + ".cover", "-d", bamFileName, "--ref_genome", faRefFile, "-t",
        gfServer.getTwoBitFile(), "--blatserver", "localhost", "--blatport", "" + gfServer.getPortNum());
    if (debug) {
      RunSystemCommand.run(extractSClipProcess, System.out, "CREST", System.out, "CREST", debug);
      RunSystemCommand.run(crestProcess, System.out, "CREST", System.out, "CREST", debug);
    } else {
      RunSystemCommand.run(extractSClipProcess, null, null, null, null, debug);
      RunSystemCommand.run(crestProcess, null, null, null, null, debug);
      new File(crestPrefix + ".sclip.txt").delete();
      new File(crestPrefix + ".cover").delete();
    }
  }

  private ArrayList<String> getBreakpoints(String crestPredFile) throws FileNotFoundException {
    Scanner sc = new Scanner(new File(crestPredFile));
    ArrayList<String> breakpoints = new ArrayList<String>();
    while (sc.hasNextLine()) {
      String line = sc.nextLine();
      String[] tokens = line.split("\t");
      breakpoints.add(tokens[0] + ":" + tokens[1]);
      breakpoints.add(tokens[4] + ":" + tokens[5]);
    }
    sc.close();
    return breakpoints;
  }

  private void extractRelevantReads(ArrayList<String> breakpoints, String bamFile, String faRefFile, String fqOutputFile)
      throws InterruptedException, IOException, Exception {
    String tmpOutput = prefix + ".tmp" + System.currentTimeMillis();
    if (breakpoints.size() > 0) {
      String[] gatkArgs = new String[12 + (2 * breakpoints.size())];
      int ind = 0;
      gatkArgs[ind++] = "-T";
      gatkArgs[ind++] = "PrintReads";
      gatkArgs[ind++] = "-I";
      gatkArgs[ind++] = bamFile;
      gatkArgs[ind++] = "-R";
      gatkArgs[ind++] = faRefFile;
      gatkArgs[ind++] = "--interval_padding";
      gatkArgs[ind++] = "25";
      for (int i = 0; i < breakpoints.size(); i++) {
        gatkArgs[ind++] = "--intervals";
        gatkArgs[ind++] = breakpoints.get(i);
      }
      gatkArgs[ind++] = "-o";
      gatkArgs[ind++] = tmpOutput + ".bam";
      gatkArgs[ind++] = "-l";
      gatkArgs[ind++] = "ERROR";
      if (debug) {
        gatkArgs[gatkArgs.length - 1] = "INFO";
      }
      RunGatkCommand.run(gatkArgs, debug);
      FileOutputStream output;
      output = new FileOutputStream(tmpOutput + ".sam");
      RunSystemCommand.run(new ProcessBuilder("samtools", "view", "-h", tmpOutput + ".bam"), output, null, System.out,
          "samtools", debug);
      RunSystemCommand.run(new ProcessBuilder("samtools", "view", "-f", "4", bamFile), output, null, System.out,
          "samtools", debug);
      output.close();
    } else {
      FileOutputStream output;
      output = new FileOutputStream(tmpOutput + ".sam");
      RunSystemCommand.run(new ProcessBuilder("samtools", "view", "-h", "-f", "4", bamFile), output, null, System.out,
          "samtools", debug);
      output.close();
    }
    Scanner sc = new Scanner(new File(tmpOutput + ".sam"));
    PrintWriter writer = new PrintWriter(fqOutputFile);
    while (sc.hasNextLine()) {
      String line = sc.nextLine();
      if (line.charAt(0) != '@') {
        String[] tokens = line.split("\t");
        if (BigInteger.valueOf(Long.valueOf(tokens[1]).byteValue()).testBit(6)) {
          writer.println("@" + tokens[0] + "-1");
        } else {
          writer.println("@" + tokens[0] + "-2");
        }
        if (BigInteger.valueOf(Long.valueOf(tokens[1]).byteValue()).testBit(4)) {
          writer.println(Genome.rc(tokens[9]));
        } else {
          writer.println(tokens[9]);
        }
        writer.println("+");
        if (BigInteger.valueOf(Long.valueOf(tokens[1]).byteValue()).testBit(4)) {
          writer.println(Genome.reverse(tokens[10]));
        } else {
          writer.println(tokens[10]);
        }
      }
    }
    sc.close();
    writer.close();
    new File(tmpOutput + ".bam").delete();
    new File(tmpOutput + ".bai").delete();
    new File(tmpOutput + ".sam").delete();
  }

  public void addReadGroups(String inputFile, String outputFile) throws IOException, Exception {
    String[] picardArgs = { "I=" + inputFile, "O=" + outputFile, "SO=coordinate", "RGID=" + rgId, "RGLB=" + rgLb,
        "RGPL=" + rgPl, "RGPU=" + rgPu, "RGSM=" + rgSm, "QUIET=TRUE", "VERBOSITY=ERROR" };
    if (debug) {
      picardArgs[picardArgs.length - 2] = "QUIET=FALSE";
      picardArgs[picardArgs.length - 1] = "VERBOSITY=INFO";
    }
    RunPicardCommand.run("AddOrReplaceReadGroups", picardArgs, debug);
  }

  public void localAlign(String fqFile, String faRefFile, String bwaIndexPrefix, String bamOutputFile)
      throws FileNotFoundException, InterruptedException, IOException, ExecutionException, Exception {
    String tmpOutput = prefix + ".tmp" + System.currentTimeMillis();
    FileOutputStream output = new FileOutputStream(tmpOutput + "-1.sam");
    if (debug) {
      RunSystemCommand.run(new ProcessBuilder("bwa", "bwasw", "-M", bwaIndexPrefix, fqFile), output, null, System.out,
          "BWA-SW", debug);
    } else {
      RunSystemCommand.run(new ProcessBuilder("bwa", "bwasw", "-M", bwaIndexPrefix, fqFile), output, null, null, null,
          debug);
    }
    output.close();
    // Sorting must be done before calmd, or else calmd will be very slow
    addReadGroups(tmpOutput + "-1.sam", tmpOutput + "-2.bam");
    output = new FileOutputStream(tmpOutput + "-3.bam");
    // calmd is getting stuck! I think the error stream is getting filled up
    // before it has a chance to be read from. Need concurrent threads?
    RunSystemCommand.run(new ProcessBuilder("samtools", "calmd", "-b", tmpOutput + "-2.bam", faRefFile), output, null,
        System.out, "samtools", debug);
    output.close();
    output = new FileOutputStream(bamOutputFile);
    RunSystemCommand.run(new ProcessBuilder("samtools", "view", "-b", "-F", "256", tmpOutput + "-3.bam"), output, null,
        System.out, "samtools", debug);
    output.close();
    RunSystemCommand.run(new ProcessBuilder("samtools", "index", bamOutputFile), System.out, "samtools", System.out,
        "samtools", debug);
    new File(tmpOutput + "-1.sam").delete();
    new File(tmpOutput + "-2.bam").delete();
    new File(tmpOutput + "-3.bam").delete();
  }

  // This should be cleaned up!
  // Will result in duplicate reads (albeit with different names)
  // When unmapped reads are "re-added" after BWA-SW
  public void mergeAlignments(String newAlignment, String firstAlignment, String secondAlignment)
      throws ExecutionException, IOException, InterruptedException {
    String tmpOutput = prefix + ".tmp" + System.currentTimeMillis();
    RunSystemCommand.run(new ProcessBuilder("samtools", "sort", secondAlignment, tmpOutput), System.out, "samtools",
        System.out, "samtools", debug);
    RunSystemCommand.run(new ProcessBuilder("samtools", "merge", newAlignment, firstAlignment, tmpOutput + ".bam"),
        System.out, "samtools", System.out, "samtools", debug);
    RunSystemCommand.run(new ProcessBuilder("samtools", "index", newAlignment), System.out, "samtools", System.out,
        "samtools", debug);
    new File(tmpOutput + ".bam").delete();
  }

  /**
   * Main method to run the ShearSV module.
   * 
   * @throws InterruptedException
   * @throws IOException
   * @throws Exception
   */
  public void execute() throws InterruptedException, IOException, Exception {

    // Initialize GFServer
    GFServer gfServer = new GFServer(prefix, origTwoBitRefFile, debug);

    try {

      // Start GFServer
      if (debug) {
        System.out.println("[SHEAR] Using port " + gfServer.getPortNum() + " for gfServer");
        System.out.println("[SHEAR] Starting gfServer...");
      }
      gfServer.start();
      if (debug) {
        System.out.println("[SHEAR] Waiting for gfServer to be ready...");
      }
      gfServer.waitForReady();
      if (debug) {
        System.out.println("[SHEAR] gfServer is ready!");
      }

      // Extract region + unmapped reads, if specified
      // Also extra original read group information
      if (region != null) {
        System.out.println("[SHEAR] Extracting reads from region " + region + "...");
        String tmpOutput = prefix + ".tmp" + System.currentTimeMillis();
        FileOutputStream output = new FileOutputStream(tmpOutput + ".bam");
        if (debug) {
          RunSystemCommand.run(new ProcessBuilder("samtools", "view", "-b", origBamFile, region), output, null,
              System.out, "samtools", debug);
        } else {
          RunSystemCommand.run(new ProcessBuilder("samtools", "view", "-b", origBamFile, region), output, null, null,
              null, debug);
        }
        output.close();
        System.out.println("[SHEAR] Extracting original read group information...");
        saveReadGroupInfo(tmpOutput + ".bam");
        new File(tmpOutput + ".bam").delete();
      } else {
        System.out.println("[SHEAR] Extracting original read group information...");
        saveReadGroupInfo(origBamFile);
      }

      // Extract unmapped reads
      // System.out.println("[SHEAR] Extracting Unmapped Reads...");
      //extractRelevantReads(new ArrayList<String>(), prefix + "-orig-with-rg.bam", origFaRefFile, prefix + "." + 0 + ".extracted.fq");

      // Create new alignment with locally-remapped originally-unmapped reads
      // System.out.println("[SHEAR] Realigning using BWA-SW local alignment...");
      // localAlign(prefix + "." + 0 + ".extracted.fq", origFaRefFile, origBwaIndexPrefix, prefix + "." + 0 + ".extracted.aligned.bam");
      // System.out.println("[SHEAR] Merging BWA-SW-aligned reads back to original alignment...");
      // mergeAlignments(prefix + "." + 0 + ".extracted.aligned.merged.bam", prefix + "-orig-with-rg.bam", prefix + "." + 0 + ".extracted.aligned.bam");

      // Initialize set of variants
      TreeSet<Variant> variants = new TreeSet<Variant>();

      // Predict SNPs/INDELs
      if (!svOnly) {
        System.out.println("[SHEAR] Predicting SNPs and INDELs...");
        variants = runHaplotypeCaller(prefix + "-orig-with-rg.bam", origFaRefFile, prefix + ".snpindel.vcf");
      }

      // Initial SV predictions
      System.out.println("[SHEAR] Making SV Predictions with CREST...");
      runCrest(prefix + "-orig-with-rg.bam", prefix + ".0", origFaRefFile, gfServer);
      // runCrest(prefix + "." + 0 + ".extracted.aligned.merged.bam", prefix + ".0", origFaRefFile, gfServer);

      // Look at breakpoints from SV predictions
      System.out.println("[SHEAR] Extracting Breakpoints...");
      ArrayList<String> breakpoints = getBreakpoints(prefix + ".0.predSV.txt");
      System.out.println("[SHEAR] Breakpoints:");
      boolean hasNewBreakpoints = true;
      if (breakpoints.size() > 0) {
        for (int i = 0; i < breakpoints.size(); i++) {
          System.out.println("[SHEAR]   " + breakpoints.get(i));
        }
      } else {
        System.out.println("[SHEAR]   None");
        hasNewBreakpoints = false;
      }

      // Continue iterations until no new breakpoints from SV predictions
      int ctr = 0;
      while ((ctr == 0) || hasNewBreakpoints) {
        // while (hasNewBreakpoints || (ctr == 0)) {
        ctr++;

        // Extract relevant reads and realign
        System.out.println("[SHEAR] Extracting Breakpoint and Unmapped Reads...");
        extractRelevantReads(breakpoints, prefix + "-orig-with-rg.bam", origFaRefFile, prefix + "." + ctr
            + ".extracted.fq");
        System.out.println("[SHEAR] Realigning using BWA-SW local alignment...");
        localAlign(prefix + "." + ctr + ".extracted.fq", origFaRefFile, origBwaIndexPrefix, prefix + "." + ctr
            + ".extracted.aligned.bam");
        if (!debug) {
          new File(prefix + "." + ctr + ".extracted.fq").delete();
        }

        // Make new SV predictions
        System.out.println("[SHEAR] Making SV Predictions with CREST...");
        runCrest(prefix + "." + ctr + ".extracted.aligned.bam", prefix + "." + ctr + "", origFaRefFile, gfServer);

        // Look at breakpoints from SV predictions
        System.out.println("[SHEAR] Extracting Breakpoints...");
        ArrayList<String> newBreakpoints = getBreakpoints(prefix + "." + ctr + ".predSV.txt");
        hasNewBreakpoints = false;
        for (int i = 0; i < newBreakpoints.size(); i++) {
          if (!breakpoints.contains(newBreakpoints.get(i))) {
            System.out.println("[SHEAR] New breakpoint: " + newBreakpoints.get(i));
            breakpoints.add(newBreakpoints.get(i));
            hasNewBreakpoints = true;
          }
        }

      }

      // No new breakpoints from SV predictions
      System.out.println("[SHEAR] No new breakpoints detected");
      if (!debug) {
        new File(prefix + "-orig-with-rg.bam").delete();
        new File(prefix + "-orig-with-rg.bam.bai").delete();
      }

      // Estimate heterogeneity for all variants
      System.out.println("[SHEAR] Estimating Heterogeneity...");
      ArrayList<CrestPrediction> crestPreds = getCrestPreds(prefix + "." + ctr + ".predSV.txt");
      crestPreds = processCrestPreds(prefix + "." + ctr + ".extracted.aligned.bam", origFaRefFile, crestPreds);
      if (!debug) {
        for (int i = 0; i <= ctr; i++) {
          new File(prefix + "." + i + ".predSV.txt").delete();
          new File(prefix + "." + i + ".extracted.aligned.bam").delete();
          new File(prefix + "." + i + ".extracted.aligned.bam.bai").delete();
        }
      }

      // Filter CREST results
      System.out.println("[SHEAR] Filtering Results...");
      crestPreds = filterCrestPreds(crestPreds);

      // Convert crestPreds to variants, fix this later to just use variants
      for (CrestPrediction pred : crestPreds) {
        if (pred.type.equals("DEL")) {
          Deletion del = new Deletion(pred.bp1region, pred.bp1 + 1, (pred.bp2 - pred.bp1) - 1);
          del.setHeterogeneity(pred.heterogeneityPercent);
          variants.add(del);
        } else if (pred.type.equals("INS")) {
          TandemDuplication tdup = new TandemDuplication(pred.bp1region, pred.bp2, (pred.bp1 - pred.bp2) + 1);
          tdup.setHeterogeneity(pred.heterogeneityPercent);
          variants.add(tdup);
        } else if (pred.type.equals("CTX") || pred.type.equals("ITX")) {
          Translocation xloc = new Translocation(pred.bp1region, pred.bp1, pred.bp2region, pred.bp2);
          xloc.setHeterogeneity(pred.heterogeneityPercent);
          variants.add(xloc);
          System.out.printf("[SHEAR]       Skipping ITX/CTX for SDI output "
              + "(Just SNPs, INDELs, deletions, inversions, and tandem duplications for now)\n");
        } else if (pred.type.equals("INV")) {
          Inversion inv = new Inversion(pred.bp1region, pred.bp1 + 1, (pred.bp2 - pred.bp1) - 1);
          inv.setHeterogeneity(pred.heterogeneityPercent);
          variants.add(inv);
        }
      }

      System.out.println("[SHEAR] Filtering SNPs/INDELs overlapping with SVs...");
      TreeSet<Variant> filteredVariants = new TreeSet<Variant>();
      for (Variant v : variants) {
        if ((v instanceof SNP) || v.isIndel()) {
          boolean overlaps = false;
          for (Variant sv : variants) {
            if (!(sv instanceof SNP) && !sv.isIndel() && v.adjacentTo(sv)) {
              overlaps = true;
              break;
            }
          }
          if (!overlaps) {
            filteredVariants.add(v);
          }
        } else {
          filteredVariants.add(v);
        }
      }

      System.out.println("[SHEAR] Outputing Results...");
      outputResults(filteredVariants);
      if (!debug) {
        new File(prefix + ".snpindel.vcf").delete();
        new File(prefix + ".snpindel.vcf.idx").delete();
      }

      System.out.println("[SHEAR] All Finished!");
    } catch (Exception e) {
      throw e;

    } finally {
      if (debug) {
        System.out.println("[SHEAR] Closing gfServer");
      }
      gfServer.close();
      if (!debug) {
        new File(prefix + ".gfServer.log").delete();
        new File(prefix + ".gfServer.out").delete();
      }
    }

  }
}
